#include <uart_regs.h>
#include <math.h>

#include "cyu3system.h"
#include "cyu3os.h"
#include "cyu3dma.h"
#include "cyu3error.h"
#include "cyfxuvcinmem.h"
#include "cyu3usb.h"
#include "cyu3uart.h"
#include "cyu3utils.h"
#include "cyu3spi.h"

/********************************************************************************
 *               Variable Declarations                                          *
 ********************************************************************************/

#ifdef USB_UVC_CLASS

/*UVC uncompressed test*/
static CyU3PDmaBuffer_t UvcdmaBuffer[2];

/* Setup data field : Request */
#define CY_U3P_USB_REQUEST_MASK                       (0x0000FF00)
#define CY_U3P_USB_REQUEST_POS                        (8)

CyU3PThread uvcAppThread;           /* Thread structure */

/* UVC Header */
uint8_t glUVCHeader[CY_FX_UVC_MAX_HEADER] =
{
    0x0C,                           /* Header Length */
    0x8C,                           /* Bit field header field */
    0x00,0x00,0x00,0x00,            /* Presentation time stamp field */
    0x00,0x00,0x00,0x00,0x00,0x00   /* Source clock reference field */
};

/* Video Probe Commit Control */
uint8_t glCommitCtrl[CY_FX_UVC_MAX_PROBE_SETTING_ALIGNED] __attribute__ ((aligned (32)));

CyU3PDmaChannel          glChHandleUVCStream;            /* DMA Channel Handle  */
static volatile CyBool_t glIsUVCApplnActive = CyFalse;   /* Whether the loopback application is active or not. */
static volatile CyBool_t glIsUVCDevConfigured = CyFalse; /* Whether SET_CONFIG is complete or not. */

#endif

#ifdef USB_UAC_CLASS

static CyU3PDmaBuffer_t UACDmaBuffer[2];

/* Setup data field : Request */
#define CY_U3P_USB_REQUEST_MASK                         (0x0000FF00)
#define CY_U3P_USB_REQUEST_POS                          (8)

/* Iscochronous Transfer length */
#define CY_FX3_ISO_XFER_LEN                             (96)
/* Give a timeout value of 5s for any flash programming. */
#define CY_FX_USB_SPI_TIMEOUT                           (5000)
/* Size in pages of the Audio sample to be streamed from the SPI Flash. */
#define CY_FX3_AUDIO_SAMPLE_SIZE_IN_PAGES               (0x3CD)

CyU3PThread uacAppThread;                              /* Thread structure */

static volatile CyBool_t glIsUACApplnActive = CyFalse; /* Whether the UAC application is active or not. */
static volatile CyBool_t glIsDevConfigured = CyFalse;  /* Whether the USB device configuration is complete. */
CyU3PDmaChannel glUacStreamHandle;                     /* DMA Channel Handle  */

#endif


#ifdef USB_CDC_CLASS

CyU3PThread       USBUARTAppThread;
CyU3PDmaChannel   glChHandleUsb2UartRecv;              /* DMA MANUAL receive channel handle.*/
CyU3PDmaChannel   glChHandleUsb2UartTrans;             /* DMA MANUAL transmit channel handle.*/
CyBool_t          glIsApplnActive = CyFalse;           /* Whether the application is active or not. */
CyU3PUartConfig_t glUartConfig = {0};                  /* Current UART configuration. */
uint8_t *USB2UARTResp;

/* CDC Class specific requests to be handled by this application. */
#define SET_LINE_CODING        0x20
#define GET_LINE_CODING        0x21
#define SET_CONTROL_LINE_STATE 0x22

#define FRAME_SIZE 				64

#endif


/* Callback to handle the USB Setup Requests and UVC Class events */
static CyBool_t
CyFxUVCApplnUSBSetupCB (
        uint32_t setupdat0, /* SETUP Data 0 */
        uint32_t setupdat1  /* SETUP Data 1 */
    )
{
	//CyU3PDebugPrint (4, "\rUVC CyFxUVCApplnUSBSetupCB\r\n");

    uint16_t readCount = 0;
    uint8_t  bRequest, bReqType;
    uint8_t  bType, bTarget, temp;
    uint16_t wValue, wIndex;
    CyBool_t isHandled = CyFalse;
    CyU3PReturnStatus_t status = CY_U3P_SUCCESS;

    /* Fast enumeration is used. Only requests addressed to the interface, class,
     * vendor and unknown control requests are received by this function. */

    /* Decode the fields from the setup request. */
    bReqType =  (setupdat0 & CY_U3P_USB_REQUEST_TYPE_MASK);                       /* Direction, type and recipient of request */
    bRequest = ((setupdat0 & CY_U3P_USB_REQUEST_MASK) >> CY_U3P_USB_REQUEST_POS); /* Specific request */
    wValue   = ((setupdat0 & CY_U3P_USB_VALUE_MASK)   >> CY_U3P_USB_VALUE_POS);   /* Any value associated with request */
    wIndex   = ((setupdat1 & CY_U3P_USB_INDEX_MASK)   >> CY_U3P_USB_INDEX_POS);   /* Any index associated with request / addressed device */

    bType    =  (bReqType  & CY_U3P_USB_TYPE_MASK);   // Type of request
    bTarget  =  (bReqType  & CY_U3P_USB_TARGET_MASK); // Recipient of request

#if 0
	CyU3PDebugPrint(4, "\rUVC bReqType: 0x%x\r\n", bReqType);
    CyU3PDebugPrint(4, "\rUVC bType:    0x%x\r\n", bType);
	CyU3PDebugPrint(4, "\rUVC bTarget:  0x%x\r\n", bTarget);
	CyU3PDebugPrint(4, "\rUVC bRequest: 0x%x\r\n", bRequest);
	CyU3PDebugPrint(4, "\rUVC wValue:   0x%x\r\n", wValue);
	CyU3PDebugPrint(4, "\rUVC wIndex:   0x%x\r\n", wIndex);
#endif

	/* Log warning if target is not an interface.
	 * Return as not handled as this device currently only supports interface requests.
	 * Initial firmware had checks on each request if
	 */
	if (bTarget != CY_U3P_USB_TARGET_INTF){
		CyU3PDebugPrint(4, "\rWarning: UVC RQT targets %d, not 1", bTarget);
		return CyFalse;
	}

	if (bType == CY_U3P_USB_STANDARD_RQT)														/*add 20210818,0x00*/
	{
		CyU3PDebugPrint(4, "USB RQT: %x %x %x %x %x\r\n", bTarget, bRequest, CY_U3P_GET_MSB(wIndex),
						CY_U3P_GET_LSB(wIndex), wValue);

		/* Handle SET_FEATURE(FUNCTION_SUSPEND) and CLEAR_FEATURE(FUNCTION_SUSPEND)
		 * requests here. It should be allowed to pass if the device is in configured
		 * state and failed otherwise. */
		switch (bRequest)
		{
		case CY_U3P_USB_SC_SET_FEATURE:
		case CY_U3P_USB_SC_CLEAR_FEATURE:
			/* wValue of 0 indicates FUNCTION_SUSPEND (USB 3.2 spec, Table 9-7, pg333)
			 * If an invalid feature selector is chosen, stall - Provided firmware had
			 * a check for wValue == 0 and would skip this case if wValue != 0. Stalling
			 * should be OK as the spec says a stall if permitted on an invalid request. */
			if (wValue != 0)
				break;

			if (glIsUVCDevConfigured)
				CyU3PUsbAckSetup();
			else
				CyU3PUsbStall(0, CyTrue, CyFalse);

			isHandled = CyTrue;
			break;

		default:
			break;
		}
	}

	/* Check for UVC Class Requests */
	if (bType == CY_U3P_USB_CLASS_RQT)
	{
		CyU3PDebugPrint (4, "UVC RQT: %x %x %x %x %x\r\n", bTarget, bRequest, CY_U3P_GET_MSB(wIndex),
				CY_U3P_GET_LSB(wIndex), wValue);

		switch (CY_U3P_GET_LSB(wIndex))
		{
		/* Handle requests addressed to the Video Control interface. */
		case CY_FX_UVC_INTERFACE_VC:
			/* Respond to VC_REQUEST_ERROR_CODE_CONTROL and stall every other request as this example does not support
			 * any of the Video Control features */
			if ((CY_U3P_GET_MSB(wIndex) == 0x00) && (wValue == CY_FX_USB_UVC_VC_RQT_ERROR_CODE_CONTROL))
			{
				temp      = CY_FX_USB_UVC_RQT_STAT_INVALID_CTRL;
				isHandled = CyTrue;
				CyU3PUsbSendEP0Data (0x01, &temp);
			}
			break;

		/* Handle requests addressed to the Video Streaming interface. */
		case CY_FX_UVC_INTERFACE_VS:
			isHandled = CyTrue;

			switch (wValue)
			{
			/* As we have a single setting, we treat both PROBE and COMMIT control requests in the same way.
			 * The commit controls sent down by the host are ignored, as we can only stream with a single
			 * setting.
			 */
			case CY_FX_USB_UVC_VS_PROBE_CONTROL:
			case CY_FX_USB_UVC_VS_COMMIT_CONTROL:
				{
					switch (bRequest)
					{
					/* We only have one functional setting. Keep returning the same as current, default
					 * minimum and maximum. */
					case CY_FX_USB_UVC_GET_CUR_REQ:
					case CY_FX_USB_UVC_GET_DEF_REQ:
					case CY_FX_USB_UVC_GET_MIN_REQ:
					case CY_FX_USB_UVC_GET_MAX_REQ:
						status = CyU3PUsbSendEP0Data(CY_FX_UVC_MAX_PROBE_SETTING,
								(uint8_t *)glProbeCtrl);
						if (status != CY_U3P_SUCCESS)
						{
							CyU3PDebugPrint(4, "CyU3PUsbSendEP0Data, error code = %d\n", status);
						}
						break;

					case CY_FX_USB_UVC_SET_CUR_REQ:
						/* Disable the low power entry to optimize USB throughput */
						CyU3PUsbLPMDisable();

						/* Read the data out into a local buffer. We do not use this data in any way. */
						status = CyU3PUsbGetEP0Data(CY_FX_UVC_MAX_PROBE_SETTING_ALIGNED,
								glCommitCtrl, &readCount);
						if (status != CY_U3P_SUCCESS)
						{
							CyU3PDebugPrint(4, "CyU3PUsbGetEP0Data failed, error code = %d\n", status);
						}

						/* Check the read count. Expecting a count of CY_FX_UVC_MAX_PROBE_SETTING bytes. */
						if (readCount != (uint16_t)CY_FX_UVC_MAX_PROBE_SETTING)
						{
							CyU3PDebugPrint(4, "Invalid number of bytes received in SET_CUR Request");
						}
						break;

					default:
						CyU3PUsbStall(0, CyTrue, CyFalse);
						break;
					}
				}
				break;

			default:
				CyU3PUsbStall(0, CyTrue, CyFalse);
				break;
			}
			break;

		/* Don't try to stall the endpoint if we have already attempted data transfer. */
		default:
			break;
		}
	}

	return isHandled;
}
